"""tables for the soil-water module"""

from ..core.tablevalidation import BaseModel
from pandera.typing import Series
import pandera as pa
from typing import Optional

class INIPRESSUREHEAD(BaseModel):
    """Initial pressure head [cm, R] as a function of soil layer [1..N, I].

    Attributes:
        ZI: Series[int]: soil depth [-1.d5..0 cm, R].
        H: Series[float]: Initial soil water pressure head [-1.d10..1.d4 cm, R].
    """
    
    ZI: Series[int] = pa.Field(ge=-1.0e5, le=0.0)
    H: Series[float] = pa.Field(ge=-1.0e10, le=1.0e4)


class MXPONDTB(BaseModel):
    """minimum thickness for runoff PONDMXTB [0..1000 cm, R] as function of time
    
    Attributes:
        DATEPMX: Series[pa.DateTime]: Date of the ponding threshold for runoff.
        PONDMXTB: Series[float]: Minimum thickness for runoff.
    """
    DATEPMX: Series[pa.DateTime]
    PONDMXTB: Series[float]


class SOILPROFILE(BaseModel):
    """Vertical discretization of soil profile
    
    Attributes:
        ISUBLAY: Series[int]: number of sub layer, start with 1 at soil surface [1..MACP, I].
        ISOILLAY: Series[int]: number of soil physical layer, start with 1 at soil surface [1..MAHO, I].
        HSUBLAY: Series[float]: height of sub layer [0..1.d4 cm, R].
        HCOMP: Series[float]: height of compartments in the sub layer [0.0..1000.0 cm, R].
        NCOMP: Series[int]: number of compartments in the sub layer (Mind NCOMP = HSUBLAY/HCOMP) [1..MACP, I].
    """

    ISUBLAY: Series[int] = pa.Field(ge=1)
    ISOILLAY: Series[int] = pa.Field(ge=1)
    HSUBLAY: Series[float] = pa.Field(ge=0.0, le=1.0e4)
    HCOMP: Series[float] = pa.Field(ge=0.0, le=1.0e3)
    NCOMP: Series[int] = pa.Field(ge=1)

class SOILHYDRFUNC(BaseModel):
    """Soil hydraulic functions table.
    
    !!! warning
        ALFAW required only when the hysteresis option is set to 1 or 2. This column is set as optional column and (for now) is not checked.

    Attributes:
        ORES (Series[float]): Residual water content [0..1 cm3/cm3, R]
        OSAT (Series[float]): Saturated water content [0..1 cm3/cm3, R]
        ALFA (Series[float]): Parameter alfa of main drying curve [0.0001..100 /cm, R]
        NPAR (Series[float]): Parameter n [1.001..9 -, R]
        LEXP (Series[float]): Exponent in hydraulic conductivity function [-25..25 -, R]
        KSATFIT (Series[float]): Fitting parameter Ksat of hydraulic conductivity function [1.d-5..1d5 cm/d, R]
        H_ENPR (Series[float]): Air entry pressure head [-40.0..0.0 cm, R]
        KSATEXM (Series[float]): Measured hydraulic conductivity at saturated conditions [1.d-5..1d5 cm/d, R]
        BDENS (Series[float]): Dry soil bulk density [100..1d4 mg/cm3, R]
        ALFAW (Optional[Series[float]]): Alfa parameter of main wetting curve in case of hysteresis [0.0001..100 /cm, R]
    """

    ORES: Series[float] = pa.Field(ge=0.0, le=1.0)
    OSAT: Series[float] = pa.Field(ge=0.0, le=1.0)
    ALFA: Series[float] = pa.Field(ge=0.0001, le=100.0)
    NPAR: Series[float] = pa.Field(ge=1.001, le=9.0)
    LEXP: Series[float] = pa.Field(ge=-25.0, le=25.0)
    KSATFIT: Series[float] = pa.Field(ge=1.0e-5, le=1.0e5)
    H_ENPR: Series[float] = pa.Field(ge=-40.0, le=0.0)
    KSATEXM: Series[float] = pa.Field(ge=1.0e-5, le=1.0e5)
    BDENS: Series[float] = pa.Field(ge=100.0, le=1.0e4)
    ALFAW: Optional[Series[float]] = pa.Field(ge=0.0001, le=100.0)


"""Commonly used ranges of values for pydantic Field() objects.

Other parameters:
    UNITRANGE (dict): Range of values between 0.0 and 1.0.
    YEARRANGE (dict): Range of values for year (0 <= x <= 366).
    DVSRANGE (dict): Range of values for development stage (0 <= x <= 2).
"""

UNITRANGE = {'ge': 0.0, 'le': 1.0}
"""Range of values between 0.0 and 1.0."""

YEARRANGE = {'ge': 0, 'le': 366}
"""Range of values for year (0 <= x <= 366)."""

DVSRANGE = {'ge': 0.0, 'le': 2.0}
"""Range of values for development stage (0 <= x <= 2)."""


"""
# Drainage subpackage

Handling drainage settings and creation of the .dra file.

Modules:
    drainage: Holds the drainage settings for the SWAP model.
    drafile: Create the .dra file.
"""

from .drainage import Drainage
from .drafile import (DraFile, DrainageFluxTable, DrainageFormula,
                      DrainageInfiltrationResitance, DraSettings)



from sqlalchemy.orm import state, session

from models import ModelRun, SWAPModel
from sqlalchemy import event, text, func, bindparam, select, insert
from sqlalchemy.orm import Session


@event.listens_for(ModelRun, 'before_insert')
def auto_set_iteration(mapper, connection, target):

    query = select(func.max(ModelRun.iteration)).where(ModelRun.model_id == bindparam('model_id')).\
        params(model_id=target.model_id)

    result = connection.execute(query).scalar()

    max_iteration = result if result is not None else 0
    target.iteration = max_iteration + 1


# This would be a more elegant solution, but it doesn't work. There might be a something wrong with the way
# I use session.

# @event.listens_for(SWAPModel, 'after_insert')
# def auto_create_iteration(mapper, connection, target):
#     # Prepare the insert statement
#     stmt = insert(ModelIteration).values(
#         model_id=target.id,
#         iteration=1,
#         swp_change={},
#         drainage_change={},
#         crop_change={},
#         met_change=None
#     )
#
#     # Execute the statement
#     connection.execute(stmt)



"""The main model class.

Classes:
    Model: Main class that runs the SWAP model.
"""

from ..core import PySWAPBaseModel
from ..core import open_file
from typing import Optional, Any
from pathlib import Path
import shutil
import tempfile
import subprocess
import os
from importlib import resources
from pandas import read_csv, to_datetime
from numpy import nan
from ..extras import HeatFlow, SoluteTransport
from ..atmosphere import Meteorology
from ..irrigation import FixedIrrigation
from ..soilwater import SoilMoisture, SnowAndFrost, Evaporation, SoilProfile, SurfaceFlow
from ..simsettings import Metadata, GeneralSettings, RichardsSettings
from ..boundary import BottomBoundary
from ..drainage import Drainage
from ..plant import Crop
from .result import Result
import warnings
import platform
from pydantic import Field

IS_WINDOWS = platform.system() == 'Windows'


class Model(PySWAPBaseModel):
    """Main class that runs the SWAP model.

    The attributes must be valid pySWAP classes. For avoiding validation errors,
    for now the attributes are defined as Any.

    Attributes:
        metadata (Any): Metadata of the model.
        general_settings (Any): Simulation settings.
        meteorology (Any): Meteorological data.
        crop (Any): Crop data.
        fixedirrigation (Any): Fixed irrigation settings.
        soilmoisture (Any): Soil moisture data.
        surfaceflow (Any): Surface flow data.
        evaporation (Any): Evaporation data.
        soilprofile (Any): Soil profile data.
        snowandfrost (Optional[Any]): Snow and frost data. Default is `SnowAndFrost(swsnow=0, swfrost=0)`.
        richards (Optional[Any]): Richards data.
        lateraldrainage (Any): Lateral drainage data.
        bottomboundary (Any): Bottom boundary data.
        heatflow (Optional[Any]): Heat flow data.
        solutetransport (Optional[Any]): Solute transport data.

    Methods:
        write_swp: Write the .swp input file.
        _copy_executable: Copy the appropriate SWAP executable to the temporary directory.
        _run_swap: Run the SWAP executable.
        _read_output: Read the output file.
        _read_output_tz: Read the output file with time zone.
        _read_vap: Read the .vap output file.
        _write_inputs: Write the input files.
        _identify_warnings: Identify warnings in the log file.
        _raise_swap_warning: Raise a warning.
        _read_output_old: Save the old output files.
        run: Run the model.
    """

    metadata: Metadata
    version: str = Field(exclude=True, default='base')
    general_settings: GeneralSettings
    meteorology: Meteorology
    crop: Crop
    fixedirrigation: FixedIrrigation
    soilmoisture: SoilMoisture
    surfaceflow: SurfaceFlow
    evaporation: Evaporation
    soilprofile: SoilProfile
    snowandfrost: Optional[SnowAndFrost] = SnowAndFrost(swsnow=0, swfrost=0)
    richards: Optional[RichardsSettings] = RichardsSettings(
        swkmean=1, swkimpl=0)
    lateraldrainage: Drainage
    bottomboundary: BottomBoundary
    heatflow: Optional[HeatFlow] = HeatFlow(swhea=0)
    solutetransport: Optional[SoluteTransport] = SoluteTransport(swsolu=0)

    def write_swp(self, path: str) -> None:
        """Write the .swp input file."""

        string = self._concat_sections()
        self.save_element(string=string, path=path,
                          filename='swap', extension='swp')
        print('swap.swp saved.')

    @staticmethod
    def _copy_executable(tempdir: Path):
        """Copy the appropriate SWAP executable to the temporary directory."""
        if IS_WINDOWS:
            exec_path = resources.files(
                "pyswap.libs.swap420-exe").joinpath("swap.exe")
            shutil.copy(str(exec_path), str(tempdir))
            print('Copying the windows version of SWAP into temporary directory...')
        else:
            exec_path = resources.files(
                "pyswap.libs.swap420-linux").joinpath("swap420")
            shutil.copy(str(exec_path), str(tempdir))
            print('Copying linux executable into temporary directory...')

    def _write_inputs(self, path: str) -> None:
        print('Preparing files...')
        self.write_swp(path)
        if self.lateraldrainage.drafile:
            self.lateraldrainage.write_dra(path)
        if self.crop.cropfiles:
            self.crop.write_crop(path)
        if self.meteorology.metfile:
            self.meteorology.write_met(path)
        if self.fixedirrigation.irgfile:
            self.irrigation.fixedirrig.write_irg(path)

    @staticmethod
    def _run_swap(tempdir: Path) -> str:
        """Run the SWAP executable.

        I do not decode the sterror because the SWAP executable
        writes errors to stdout. It will be easy to implement reading
        stderr later if needed.

        Returns:
            str: stdout.
        """

        swap_path = Path(tempdir, 'swap.exe') if IS_WINDOWS else './swap420'

        p = subprocess.Popen(swap_path,
                             stdout=subprocess.PIPE,
                             stdin=subprocess.PIPE,
                             stderr=subprocess.STDOUT,
                             cwd=tempdir)

        stdout = p.communicate(input=b'\n')[0]

        return stdout.decode()

    @staticmethod
    def _read_output(path: Path):
        """Read the output csv file."""
        df = read_csv(path, comment='*', index_col='DATETIME')
        df.index = to_datetime(df.index)

        return df

    @staticmethod
    def _read_output_tz(path: Path):
        """Read the output csv file with the depth information."""
        df = read_csv(path, comment='*', index_col='DATE')
        df.index = to_datetime(df.index)

        return df

    @staticmethod
    def _read_vap(path: Path):
        df = read_csv(path, skiprows=11, encoding_errors='replace')
        df.columns = df.columns.str.strip()
        df.replace(r'^\s*$', nan, regex=True, inplace=True)
        return df

    def _read_log_file(self, directory: Path) -> str:
        """Read the log file."""
        log_files = [f for f in Path(directory).glob(
            '*.log') if f.name != 'reruns.log']

        if len(log_files) == 0:
            raise FileNotFoundError("No .log file found in the directory.")
        elif len(log_files) > 1:
            raise FileExistsError(
                "Multiple .log files found in the directory.")

        log_file = log_files[0]

        with open(log_file, 'r') as file:
            log_content = file.read()

        return log_content

    @staticmethod
    def _identify_warnings(log: str) -> list[Warning]:
        """Read through the log file and catch warnings emitted by the SWAP executable."""
        lines = log.split('\n')
        warnings = [line for line in lines
                    if line.strip().lower().startswith('warning')]

        return warnings

    def _raise_swap_warning(self, message):
        warnings.warn(message, Warning, stacklevel=3)

    def _read_output_old(self, tempdir: Path):
        """Read all output files that are not in csv format as strings."""
        list_dir = os.listdir(tempdir)
        list_dir = [f for f in list_dir if not f.find(
            self.general_settings.outfil) and not f.endswith('.csv')]

        if list_dir:
            dict_files = {f.split('.')[1]: open_file(Path(tempdir, f))
                          for f in list_dir}

        return dict_files

    def run(self, path: str | Path, silence_warnings: bool = False, old_output: bool = False):
        """Main function that runs the model.

        Parameters:
            path (str): Path to the working directory.
            silence_warnings (bool): If True, warnings will not be printed.
            old_output (bool): If True, the old output files (like .vap) will be saved to a dictionary.

        !!! todo

            It would be nice to have a nice output string that will concatenate all output
            including warnings and/or errors.


        !!! warning

            Reruns are for now not supported. Multiple runs of the model can be achieved by running
            model.run() multiple times.
        """
        with tempfile.TemporaryDirectory(dir=path) as tempdir:

            self._copy_executable(tempdir)
            self._write_inputs(tempdir)

            result = self._run_swap(tempdir)

            if 'normal completion' not in result:
                raise Exception(
                    f'Model run failed. \n {result}')

            print(result)

            log = self._read_log_file(tempdir)
            warnings = self._identify_warnings(log)

            if warnings and not silence_warnings:
                print('Warnings:')
                for warning in warnings:
                    self._raise_swap_warning(message=warning)

            if old_output:
                dict_files = self._read_output_old(tempdir)

            result = Result(
                output=self._read_output(
                    Path(tempdir, f'{self.general_settings.outfil}_output.csv')) if self.general_settings.inlist_csv else None,
                output_tz=self._read_output_tz(
                    Path(tempdir, f'{self.general_settings.outfil}_output_tz.csv')) if self.general_settings.inlist_csv_tz else None,
                log=log,
                output_old=dict_files if old_output else None,
                warning=warnings
            )

            return result



from ..core.tablevalidation import BaseModel
from pandera.typing import Series
import pandera as pa

class DAILYMETEODATA(BaseModel):
    Station: Series[str]
    DD: Series[str]
    MM: Series[str]
    YYYY: Series[str]
    RAD: Series[float]
    Tmin: Series[float]
    Tmax: Series[float]
    HUM: Series[float]
    WIND: Series[float]
    RAIN: Series[float]
    ETref: Series[float]
    WET: Series[float]

class SHORTINTERVALMETEODATA(BaseModel):
    Date: Series[pa.DateTime]
    Record: Series[int] = pa.Field(ge=1, le=10)
    Rad: Series[float]
    Temp: Series[float]
    Hum: Series[float]
    Wind: Series[float]
    Rain: Series[float]

class DETAILEDRAINFALL(BaseModel):
    Station: Series[str]
    Day: Series[int]
    Month: Series[int]
    Year: Series[int]
    Time: Series[float]
    Amount: Series[float]


"""
# Atmosphere subpackage

Meteorological settings and data for SWAP simulations.

Modules:
    metfile: meteorological data for SWAP simulations
    meteorology: meteorological settings for SWAP simulations
___________________________________________________________________________
"""

from .meteorology import *
from .metfile import *



"""Create the irrigation file for the SWAP model.

Classes:
    IrrigationFile: The irrigation file.
"""

from pydantic import Field
from pandas import read_csv
from ..core import PySWAPBaseModel
from pandas import DataFrame


class IrgFile(PySWAPBaseModel):
    """The irrigation file.

    !!! warning
        The irrigation file is the first to have pandera validation. However, 
        it is not yet complete. Some columns are set to non-required, but they
        might be required if solute transport is used.

    Attributes:
        irgfil (str): the name of the irgfile without .irg extension.
        content (DataFrame): The content of the irrigation file.
    """

    irgfil: str
    content: DataFrame = Field(exclude=True)


def irg_from_csv(irgfil: str, path: str) -> IrgFile:
    """Load the irrigation file from a CSV file.

    Parameters:
        irgfil (str): the name of the irgfile without .irg extension.
        path (str): The path to the CSV file.

    Returns:
        IrgFile: The irrigation file.
    """
    return IrgFile(content=read_csv(path), irgfil=irgfil)



"""Default values for the Richards' equation"""

DERAULT_RICHARDS = {
    'dtmin': 1.0e-7,
    'dtmax': 0.1,
    'gwlconv': 20.0,
    'critdevh1cp': 1.0e-2,
    'critdevh2cp': 1.0e-1,
    'critdevponddt': 1.0e-4,
    'maxit': 30,
    'maxbacktr': 10,
    'swkmean': 1,
    'swkimpl': 0
}



"""
Plotting functionality for pySWAP.

Modules:
    evapotranspiration: Functions for plotting evapotranspiration data.
    gwl: Functions for plotting groundwater level data.
    watercontent: Functions for plotting water content data.
"""
from .evapotranspiration import *
from .gwl import *
from .watercontent import *



"""
SQLAlchemy models for the pySWAP database.

Important assumptions:
    1. There is always an initial SWAP model which is saved in the SWAPModel table.
    2. Each SWAPModel belongs to only one project.
    3. Each SWAPModel has one or mode runs which are saved in the ModelRun table.
    4. One ModelRun has exactly one ModelOutput
"""

from sqlalchemy import Column, Integer, String, ForeignKey, LargeBinary, UniqueConstraint, Table, PrimaryKeyConstraint
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.sqlite import JSON

Base = declarative_base()


data_swaprun_association = Table(
    'data_swaprun_association',
    Base.metadata,
    Column('data_id', Integer, ForeignKey('data.did')),
    Column('swaprun_mid', String(50)),
    Column('swaprun_rid', String(50)),
)


class Data(Base):
    """Stores data files like meteo data or crop data."""

    __tablename__ = 'data'

    did = Column(Integer, primary_key=True)
    fname = Column(String(50), nullable=False, unique=True)
    datafile = Column(LargeBinary)
    swapruns = relationship(
        'SWAPRun', secondary=data_swaprun_association, back_populates='data')


class SWAPRun(Base):
    """This model stores sections of the .swp file"""

    __tablename__ = 'swapmodel'

    mid = Column(String(50), primary_key=True)
    rid = Column(String(50), default='initial', primary_key=True)
    swp = Column(LargeBinary, nullable=True)
    data = relationship(
        'Data', secondary=data_swaprun_association, back_populates='swapruns')
    result = Column(LargeBinary, nullable=True)

    __table_args__ = (UniqueConstraint('mid', 'rid', name='uq_model_run'),)


def test_insert_and_retrieve(session):
    # Create Data objects
    data1 = Data(fname='data1', datafile=b'datafile1')
    data2 = Data(fname='data2', datafile=b'datafile2')
    data3 = Data(fname='data3', datafile=b'datafile3')

    # Create SWAPRun objects
    swaprun1 = SWAPRun(mid='model1', rid='1', swp=b'swp1')
    swaprun2 = SWAPRun(mid='model1', rid='2', swp=b'swp2')
    swaprun3 = SWAPRun(mid='model2', rid='1', swp=b'swp3')

    # Associate Data objects with SWAPRun objects
    swaprun1.data = [data1, data2]
    swaprun2.data = [data2, data3]
    swaprun3.data = [data3]

    # Add objects to the session
    session.add_all([data1, data2, data3, swaprun1, swaprun2, swaprun3])
    session.commit()

    # Retrieve all SWAPRun objects along with related Data objects
    all_swapruns = session.query(SWAPRun).options(
        relationship(Data, secondary=data_swaprun_association)).all()

    # Print retrieved objects
    for swaprun in all_swapruns:
        print("SWAPRun:", swaprun.mid, swaprun.rid)
        for data in swaprun.data:
            print("  Data:", data.fname)



"""Heat flow settings for SWAP simulation.

Classes:
    HeatFlow: Heat flow settings for SWAP simulation.
"""
from ..core import PySWAPBaseModel
from ..core import Table
from typing import Literal, Optional
from typing_extensions import Self
from pydantic import model_validator


class HeatFlow(PySWAPBaseModel):
    """Heat flow settings for SWAP simulation.

    !!! warning

        table_initsoil is not validated because it depends of swinco setting.

    Attributes:
        swhea (Literal[0, 1]): Switch for heat flow.
        swcalt (Optional[Literal[1, 2]]):

            * 1 - analytical method
            * 2 - numerical method

        tampli (Optional[float]): Amplitude of annual temperature wave at soil surface [0..50 oC, R]
        tmean (Optional[float]): Mean annual temperature at soil surface [-10..30 oC, R]
        timref (Optional[float]): Time at which the sinus temperature wave reaches it's top [0..366.0 d, R]
        ddamp (Optional[float]): Damping depth of soil temperature wave [1..500 cm, R]
        swtopbhea (Optional[Literal[1, 2]]): Define top boundary condition

            * 1 - use air temperature of meteo input file as top boundary
            * 2 - use measured top soil temperature as top boundary

        tsoilfile (Optional[str]): name of input file with soil surface temperatures without extension .TSS
        swbotbhea (Optional[Literal[1, 2]]): Define bottom boundary condition

            * 1 - no heat flux
            * 2 - prescribe bottom temperature

        table_soiltextures (Optional[Table]): for each physical soil layer the soil texture (g/g mineral parts) and the organic matter content (g/g dry soil)
        table_initsoil (Optional[Table]): initial temperature TSOIL [-50..50 oC, R] as function of soil depth ZH [-100000..0 cm, R]
        table_bbctsoil (Optional[Table]): bottom boundary temperature TBOT [-50..50 oC, R] as function of date DATET [date]
    """

    swhea: Literal[0, 1]
    swcalt: Optional[Literal[1, 2]] = None
    tampli: Optional[float] = None
    tmean: Optional[float] = None
    timref: Optional[float] = None
    ddamp: Optional[float] = None
    swtopbhea: Optional[Literal[1, 2]] = None
    tsoilfile: Optional[str] = None
    swbotbhea: Optional[Literal[1, 2]] = None
    table_soiltextures: Optional[Table] = None
    table_initsoil: Optional[Table] = None
    table_bbctsoil: Optional[Table] = None

    @model_validator(mode='after')
    def _check_heatflow(self) -> Self:
        if self.swhea == 1:
            assert self.swcalt is not None, "swcalt must be specified if swhea is 1"
            if self.swcalt == 1:
                assert self.tampli is not None, "tampli must be specified if swcalt is 1"
                assert self.tmean is not None, "tmean must be specified if swcalt is 1"
                assert self.timref is not None, "timref must be specified if swcalt is 1"
                assert self.ddamp is not None, "ddamp must be specified if swcalt is 1"
            elif self.swcalt == 2:
                assert self.table_soiltextures is not None, "table_soiltextures must be specified if swcalt is 2"
                if self.swtopbhea == 2:
                    assert self.tsoilfile is not None, "tsoilfile must be specified if swtopbhea is 2"
                if self.swbotbhea == 2:
                    assert self.table_bbctsoil is not None, "table_bbctsoil must be specified if swbotbhea is 2"
        return self



"""
# Soil-water subpackage

Settings related to soil and water.

Modules:
    evaporation: Evaporation settings.
    soilmoisture: Soil moisture settings.
    soilprofile: Soil profile settings.
    surfaceflow: Surface flow settings.
    snow: Snow and frost settings.
"""

from .snow import SnowAndFrost
from .evaporation import Evaporation
from .soilmoisture import SoilMoisture
from .soilprofile import SoilProfile
from .surfaceflow import SurfaceFlow
from .tables import INIPRESSUREHEAD, MXPONDTB, SOILPROFILE, SOILHYDRFUNC



"""Tables for the irrigation settings.

Classes:

"""
from ..core.tablevalidation import BaseModel
from pandera.typing import Series
import pandera as pa


class IRRIGATION(BaseModel):
    """information for each fixed irrigation event.

    Attributes:
        IRDATE (Series[datetime]):date of irrigation.
        IRDEPTH (Series[float]): amount of water [0..1000 mm, R].
        IRCONC (Series[float]): concentration of irrigation water [0..1000 mg/cm3, R].
        IRTYPE (Series[int]): type of irrigation 

            * 0 - sprinkling 
            * 1 - surface

    """

    IRDATE: Series[pa.DateTime]
    IRDEPTH: Series[float] = pa.Field(ge=0.0, le=1000.0)
    IRCONC: Series[float] = pa.Field(ge=0.0, le=1000.0)
    IRTYPE: Series[int] = pa.Field(ge=0, le=1)



"""
Metadata class collecting basic information about the model.

Classes:
    Metadata: Metadata of a SWAP model.
"""

from ..core import PySWAPBaseModel
from typing import Optional
from pydantic import Field


class Metadata(PySWAPBaseModel):
    """Metadata of a SWAP model.

    Attributes:
        author (str): Author of the model.
        institution (str): Institution of the author.
        email (str): Email of the author.
        project (str): Name of the project.
        swap_ver (str): Version of SWAP used.
        comment (Optional[str]): Comment about the model.
    """

    author: str = Field(exclude=True)
    institution: str = Field(exclude=True)
    email: str = Field(exclude=True)
    project: str
    swap_ver: str = Field(exclude=True)
    comment: Optional[str] = Field(default=None, exclude=True)



"""
# Core subpackage

Core package containing the main classes and functions for the SWAP model.

Modules:
    basemodel: Base model class for pySWAP.
    fields: Field types for pyswap used for serialization.
    files: Functions to interact with file system.
    serializers: Functions to fine tune the serializatino of pySWAP objects.
    valueranges: Objects containing value ranges used is validation of pySWAP objects.
    model: Contains the classes for the model of the simulation.
    result: Contains the classes for the results of the simulation.
___________________________________________________________________________
"""

from .basemodel import PySWAPBaseModel
from .fields import (Table, Arrays, CSVTable, DayMonth, IntList,
                     StringList, FloatList, DateList, Switch, ObjectList)
from .files import open_file, save_file
from .serializers import (serialize_table, serialize_arrays, serialize_csv_table,
                          is_scientific_notation, quote_string, serialize_object_list)
from .valueranges import UNITRANGE, YEARRANGE, DVSRANGE
from .tablevalidation import BaseModel



from datetime import date as d
from ..core import PySWAPBaseModel
from ..core import (DayMonth, DateList, StringList, FloatList)
from ..core import YEARRANGE, UNITRANGE
from typing import Literal, Optional
from typing_extensions import Self
from datetime import date as d
from pydantic import Field, model_validator
import platform

IS_WINDOWS = platform.system() == 'Windows'
BASE_PATH = '.\\' if IS_WINDOWS else './'


class GeneralSettings(PySWAPBaseModel):
    """General settings of the simulation.

    !!! todo
        * remove the individual file extension switches and replace with a list of extensions. Make the class
            automatically generate the switches based on the list of extensions.

    Attributes:
        pathwork (str): Path to the working directory
        pathatm (str): Path to folder with weather files
        pathcrop (str): Path to folder with crop files
        pathdrain (str): Path to folder with drainage files
        swscre (Literal[0, 1, 3]): Switch, display progression of simulation run to screen
        swerror (Literal[0, 1]): Switch for printing errors to screen
        tstart (d): Start date of simulation run, give day-month-year
        tend (d): End date of simulation run, give day-month-year
        nprintday (int): Number of output times during a day
        swmonth (Literal[0, 1]): Switch, output each month
        swyrvar (Literal[0, 1]): Output times for overall water and solute balances in *.BAL and *.BLC file: choose output 
            at a fixed date each year or at different dates
        period (Optional[int]): Fixed output interval
        swres (Optional[Literal[0, 1]]): Switch, reset output interval counter each year
        swodat (Optional[Literal[0, 1]]): Switch, extra output dates are given in table below
        outdatin (Optional[DateList]): list of specific dates
        datefix (Optional[DayMonth]): fixed date for output
        outdat (Optional[DateList]): specify all output dates
        outfil (str): Generic file name of output files
        swheader (Literal[0, 1]): Print header at the start of each balance period
        swwba (Literal[0, 1]): Switch, output daily water balance
        swend (Literal[0, 1]): Switch, output end-conditions
        swvap (Literal[0, 1]): Switch, output soil profiles of moisture, solute and temperature
        swbal (Literal[0, 1]): Switch, output file with yearly water balance
        swblc (Literal[0, 1]): Switch, output file with detailed yearly water balance
        swsba (Literal[0, 1]): Switch, output file of daily solute balance
        swate (Literal[0, 1]): Switch, output file with soil temperature profiles
        swbma (Literal[0, 1]): Switch, output file with water fluxes, only for macropore flow
        swdrf (Literal[0, 1]): Switch, output of drainage fluxes, only for extended drainage
        swswb (Literal[0, 1]): Switch, output surface water reservoir, only for extended drainage
        swini (Literal[0, 1]): Switch, output of initial SoilPhysParam and HeatParam
        swinc (Literal[0, 1]): Switch, output of water balance increments
        swcrp (Literal[0, 1]): Switch, output of simple or detailed crop growth model
        swstr (Literal[0, 1]): Switch, output of stress values for wetness, drought, salinity and frost
        swirg (Literal[0, 1]): Switch, output of irrigation gifts
        swcsv (Literal[0, 1]): Switch, csv output
        inlist_csv (Optional[StringList]): list of variables for the csv output
        swcsv_tz (Literal[0, 1]): Switch, csv output with depth
        inlist_csv_tz (Optional[StringList]): list of variables for the csv tz output
        swafo (Literal[0, 1, 2]): Switch, output file with formatted hydrological data
        swaun (Literal[0, 1, 2]): Switch, output file with unformatted hydrological data
        critdevmasbal (Optional[float]): Critical Deviation in water balance during PERIOD
        swdiscrvert (Literal[0, 1]): Switch to convert vertical discretization
        numnodnew (Optional[int]): New number of nodes
        dznew (Optional[FloatList]): Thickness of compartments
    """

    pathwork: str = BASE_PATH
    pathatm: str = BASE_PATH
    pathcrop: str = BASE_PATH
    pathdrain: str = BASE_PATH
    swscre: Literal[0, 1, 3] = 0
    swerror: Literal[0, 1] = 0

    tstart: d  # convert this to DD-MM-YYYY
    tend: d  # convert this to DD-MM-YYYY

    nprintday: int = Field(default=1, ge=1, le=1440)
    swmonth: Literal[0, 1] = 1
    swyrvar: Literal[0, 1] = 0
    # if swmonth is 0
    period: Optional[int] = Field(default=None, **YEARRANGE)
    swres: Optional[Literal[0, 1]] = None
    swodat: Optional[Literal[0, 1]] = None
    # if swyrvar is 1
    outdatin: Optional[DateList] = None
    datefix: Optional[DayMonth] = None
    outdat: Optional[DateList] = None

    outfil: str = "result"
    swheader: Literal[0, 1] = 0
    swwba: Literal[0, 1] = 0
    swend: Literal[0, 1] = 0
    swvap: Literal[0, 1] = 0
    swbal: Literal[0, 1] = 0
    swblc: Literal[0, 1] = 1
    swsba: Literal[0, 1] = 0
    swate: Literal[0, 1] = 0
    swbma: Literal[0, 1] = 0
    swdrf: Literal[0, 1] = 0
    swswb: Literal[0, 1] = 0
    swini: Literal[0, 1] = 0
    swinc: Literal[0, 1] = 0
    swcrp: Literal[0, 1] = 0
    swstr: Literal[0, 1] = 0
    swirg: Literal[0, 1] = 0
    swcsv: Literal[0, 1] = 1
    inlist_csv: Optional[StringList] = None
    swcsv_tz: Literal[0, 1] = 0
    inlist_csv_tz: Optional[StringList] = None
    swafo: Literal[0, 1, 2] = 0
    swaun: Literal[0, 1, 2] = 0
    critdevmasbal: Optional[float] = Field(default=None, **UNITRANGE)
    swdiscrvert: Literal[0, 1] = 0
    numnodnew: Optional[int] = None
    dznew: Optional[FloatList] = None

    @model_validator(mode='after')
    def _validate_model(self) -> Self:

        if not self.swmonth:
            assert self.period is not None, "period is required when swmonth is 0"
            assert self.swres is not None, "swres is required when swmonth is 0"
            assert self.swodat is not None, "swodat is required when swmonth is 0"
            if self.swodat:
                assert self.outdatin is not None, "outdatin is required when swodat is 1"

        if self.swyrvar:
            assert self.outdat is not None, "outdat is required when svyrvar is 1"
        else:
            assert self.datefix is not None, "datefix is required when swyrvar is 0"

        if self.swafo in [1, 2] or self.swaun in [1, 2]:
            assert self.critdevmasbal is not None, "critdevmasbal is required when SWAFO = 1 or 2 or SWAUN = 1 or 2"
            assert self.swdiscrvert, "SWDISCRVERT is required when SWAFO = 1 or 2 or SWAUN = 1 or 2"
        if self.swdiscrvert:
            assert self.numnodnew is not None, "NUMNODNEW is required when SWDISCRVERT = 1"
            assert self.dznew is not None, "DZNEW is required when SWDISCRVERT = 1"

        return self



"""
# Plant subpackage

Crop settings and files for the SWAP simulation.

Modules:
    crop: The crop settings.
    crpfile: The crop file.
    create_crp_tables: Experimental module with functions to create crop tables.
"""

from .crop import (CropFile, Crop, CropDevelopmentSettingsWOFOST, CropDevelopmentSettings,
                   CropDevelopmentSettingsFixed, CO2Correction, CompensateRWUStress, Preparation,
                   OxygenStress, DroughtStress, Interception, CropDevelopmentSettingsGrass, SaltStress, GrasslandManagement)
from .tables import (RDTB, RDCTB, GCTB, CHTB, KYTB, MRFTB, WRTB, CROPROTATION, DTSMTB,
                     SLATB, AMAXTB, TMPFTB, TMNFTB, RFSETB, FRTB, FLTB, FSTB, FOTB, RDRRTB, RDRSTB, CHTB_GRASS, DMGRZTB, LSDATB, LSDBTB, DMMOWTB, DMMOWDELAY,
                     SLATB_GRASS, AMAXTB_GRASS, RFSETB_GRASS, FRTB_GRASS, FLTB_GRASS, FSTB_GRASS, RDRRTB_GRASS, RDRSTB_GRASS, RLWTB)



"""Capturing model results.

Tip:
    The Result class is now focusing on the output in CSV format and the log file. the other
    result files are also retrieved as a list of strings which user can access if needed.

Classes:
    Result: Stores the result of a model run.

"""
from pydantic import BaseModel, ConfigDict, computed_field, Field
from typing import Optional, List, Dict
from pandas import DataFrame
import re


class Result(BaseModel):
    """Class to store the result of a model run.

    Attributes:
        log (str): The log file of the model run.
        summary (str, optional): The summary file of the model run.
        output (DataFrame, optional): The output file of the model run.
        output_tz (DataFrame, optional): The output file of the model run with timezone.
        output_old (Dict[str, str], optional): The old output files of the model run.
        warning (List[str], optional): The warnings of the model run.
        model_config (ConfigDict): The configuration for the model.

    Methods:
        iteration_stats (str): The part of the log file that describes the iteration statistics.
        blc_summary (str): The .blc file if it exists.
        water_balance (str): The water balance of the model run.
    """

    log: str
    output: Optional[DataFrame] = Field(default=None, repr=False)
    output_tz: Optional[DataFrame] = Field(default=None, repr=False)
    output_old: Optional[Dict[str, str]] = Field(default=None, repr=False)
    warning: Optional[List[str]] = Field(default=None, repr=False)

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=True,
        extra='forbid'
    )

    @computed_field(return_type=str)
    def iteration_stats(self):
        """Return the part of the string that describes the iteration statistics."""
        return re.search(r'.*(Iteration statistics\s*.*)$', self.log, re.DOTALL)[1]

    @computed_field(return_type=str)
    def blc_summary(self):
        """Return the .blc file if it exists."""
        return self.output_old.get('blc') if self.output_old else None

    def yearly_summary(self):
        """Return yearly sums of all output variables."""
        return self.output.resample('YE').sum()



"""
All modules related to the database connection and models.


!!! warning
    This module is in review and may be significantly changed or removed in the future.

!!! tip
    pySWAP is meant to enable saving model runs to a SQLite database. That way, the users will be 
    able to easily exchange models they built and run them on their own machines.
"""
from .hdf5 import HDF5



"""Solute transport settings for the SWAP simulation.

Classes:
    SoluteTransport: Solute transport settings.
"""

from ..core import PySWAPBaseModel
from ..core import Table
from typing import Literal, Optional
from typing_extensions import Self


class SoluteTransport(PySWAPBaseModel):
    """
    Solute transport settings.

    !!! warning

        Validation not yet included in the current release.

    Attributes:
        swsolu (Literal[0, 1]): 
        cpre (Optional[float]):
        cdrain (Optional[float]):
        swbotbc (Optional[Literal[0, 1]]):
        cseep (Optional[float]):
        ddif (Optional[float]):
        tscf (Optional[float]):
        swsp (Optional[Literal[0, 1]]):
        frexp (Optional[float]):
        cref (Optional[float]):
        swdc (Optional[Literal[0, 1]]):
        gampar (Optional[float]):
        rtheta (Optional[float]):
        bexp (Optional[float]):
        swbr (Optional[Literal[0, 1]]):
        daquif (Optional[float]):
        poros (Optional[float]):
        kfsat (Optional[float]):
        decsat (Optional[float]):
        cdraini (Optional[float]):
        table_cseeparrtb (Optional[Table]):
        table_inissoil (Optional[Table]):
        table_miscellaneous (Optional[Table]):
    """
    swsolu: Literal[0, 1]
    cpre: Optional[float] = None
    cdrain: Optional[float] = None
    swbotbc: Optional[Literal[0, 1]] = None
    cseep: Optional[float] = None
    ddif: Optional[float] = None
    tscf: Optional[float] = None
    swsp: Optional[Literal[0, 1]] = None
    frexp: Optional[float] = None
    cref: Optional[float] = None
    swdc: Optional[Literal[0, 1]] = None
    gampar: Optional[float] = None
    rtheta: Optional[float] = None
    bexp: Optional[float] = None
    swbr: Optional[Literal[0, 1]] = None
    daquif: Optional[float] = None
    poros: Optional[float] = None
    kfsat: Optional[float] = None
    decsat: Optional[float] = None
    cdraini: Optional[float] = None
    table_cseeparrtb: Optional[Table] = None
    table_inissoil: Optional[Table] = None
    table_miscellaneous: Optional[Table] = None
